@@ -20,12 +20,6 @@
  * Leave it commented up for normal operation and releases! */
 /* #define REPORT_UNKNOWN 1 */
 
-double sRGB_gamma=0.45455;      /* For HTML, which runs
-				 * according to sRGB standard. Number
-				 * in HTML tag is linear to photons raised
-				 * to this power.
-				 */
-
 unsigned aspect=65536; /* aspect=65536 for 320x240
 			* aspect=157286 for 640x200 (tall pixels)
 			* Defines aspect ratio of screen pixels.
@@ -48,39 +42,39 @@ unsigned aspect=65536; /* aspect=65536 for 320x240
  * If you change them, don't wonder if the letters start to look strange.
  * The numers in the comments denote which height the line applies for.
  */
-static float fancy_constants[64]={
+static const float fancy_constants[64]={
 	(float)0,(float)3,	/*  1 */
-	(float).1,(float)3,   	/*  2 */
-	(float).2,(float)3,   	/*  3 */
-	(float).3,(float)2.9,  	/*  4 */
-	(float).4,(float)2.7,  	/*  5 */
-	(float).4,(float)2.5,  	/*  6 */
-	(float).4,(float)2,   	/*  7 */
-	(float).5,(float)2, 	/*  8 */
-	(float).4,(float)2, 	/*  9 */
-	(float).38,(float)1.9, 	/* 10 */
-	(float).36,(float)1.8, 	/* 11 */
-	(float).33,(float)1.7, 	/* 12 */
-	(float).30,(float)1.6, 	/* 13 */
-	(float).25,(float)1.5, 	/* 14 */
-	(float).20,(float)1.5, 	/* 15 */
-	(float).15,(float)1.5, 	/* 16 */
-	(float).14,(float)1.5, 	/* 17 */
-	(float).13,(float)1.5, 	/* 18 */
-	(float).12,(float)1.5, 	/* 19 */
-	(float).12,(float)1.5, 	/* 20 */
-	(float).12,(float)1.5, 	/* 21 */
-	(float).12,(float)1.5, 	/* 22 */
-	(float).11,(float)1.5, 	/* 23 */
-	(float).10,(float)1.4, 	/* 24 */
-	(float).09,(float)1.3, 	/* 25 */
-	(float).08,(float)1.3, 	/* 26 */
-	(float).04,(float)1.2, 	/* 27 */
-	(float).04,(float)1.2, 	/* 28 */
-	(float).02,(float)1.1, 	/* 29 */
-	(float).02,(float)1.1, 	/* 30 */
-	(float).01,(float)1, 	/* 31 */
-	(float).01,(float)1  	/* 32 */
+	(float).1,(float)3,	/*  2 */
+	(float).2,(float)3,	/*  3 */
+	(float).3,(float)2.9,	/*  4 */
+	(float).4,(float)2.7,	/*  5 */
+	(float).4,(float)2.5,	/*  6 */
+	(float).4,(float)2,	/*  7 */
+	(float).5,(float)2,	/*  8 */
+	(float).4,(float)2,	/*  9 */
+	(float).38,(float)1.9,	/* 10 */
+	(float).36,(float)1.8,	/* 11 */
+	(float).33,(float)1.7,	/* 12 */
+	(float).30,(float)1.6,	/* 13 */
+	(float).25,(float)1.5,	/* 14 */
+	(float).20,(float)1.5,	/* 15 */
+	(float).15,(float)1.5,	/* 16 */
+	(float).14,(float)1.5,	/* 17 */
+	(float).13,(float)1.5,	/* 18 */
+	(float).12,(float)1.5,	/* 19 */
+	(float).12,(float)1.5,	/* 20 */
+	(float).12,(float)1.5,	/* 21 */
+	(float).12,(float)1.5,	/* 22 */
+	(float).11,(float)1.5,	/* 23 */
+	(float).10,(float)1.4,	/* 24 */
+	(float).09,(float)1.3,	/* 25 */
+	(float).08,(float)1.3,	/* 26 */
+	(float).04,(float)1.2,	/* 27 */
+	(float).04,(float)1.2,	/* 28 */
+	(float).02,(float)1.1,	/* 29 */
+	(float).02,(float)1.1,	/* 30 */
+	(float).01,(float)1,	/* 31 */
+	(float).01,(float)1	/* 32 */
 };
 
 
@@ -166,6 +160,7 @@ static inline void add_row_gray(unsigned *row_buf, unsigned char *ptr, int n,
 static inline void add_col_color(scale_t *col_buf, unsigned short *ptr,
 	int line_skip, int n, ulonglong weight)
 {
+	if (!weight) return;
 #ifndef USE_FP_SCALE
 	if (weight < 0x10000) {
 		unsigned short weight_16 = (unsigned short)weight;
@@ -195,6 +190,7 @@ static inline void add_col_color(scale_t *col_buf, unsigned short *ptr,
 static inline void add_row_color(scale_t *row_buf, unsigned short *ptr,
 	int n, ulonglong weight)
 {
+	if (!weight) return;
 #ifndef USE_FP_SCALE
 	if (weight < 0x10000) {
 		unsigned short weight_16 = (unsigned short)weight;
@@ -398,23 +394,25 @@ static inline ulonglong multiply_int(int a, int b)
  * Frees input and allocates output.
  * We assume unsigned holds at least 32 bits
  */
-static inline void enlarge_color_horizontal(unsigned short *ina, int ix, int y,
+static inline void enlarge_color_horizontal(unsigned short *in, int ix, int y,
 	unsigned short **outa, int ox)
 {
+#ifdef HAVE_OPENMP
+	int use_omp;
+#endif
 	scale_t *col_buf;
 	int a;
-	ulonglong total,out_pos,in_pos,in_begin,in_end;
 	unsigned skip=3*ix;
 	unsigned oskip=3*ox;
-	unsigned short *out, *in;
+	unsigned short *out;
 
-	if (!ina) {
+	if (!in) {
 		*outa = NULL;
 		return;
 	}
 
 	if (ix==ox){
-		*outa=ina;
+		*outa=in;
 		return;
 	}
 	if (ox && (unsigned)ox * (unsigned)y / (unsigned)ox != (unsigned)y) overalloc();
@@ -422,51 +420,73 @@ static inline void enlarge_color_horizontal(unsigned short *ina, int ix, int y,
 	out=mem_alloc_mayfail(sizeof(*out)*3*ox*y);
 	*outa=out;
 	if (!out) {
-		mem_free(ina);
+		mem_free(in);
 		return;
 	}
-	in=ina;
 	if (ix==1){
-		for (;y;y--,in+=3) for (a=ox;a;a--,out+=3){
-			out[0]=in[0];
-			out[1]=in[1];
-			out[2]=in[2];
+		unsigned short *inp = in;
+		for (;y;y--,inp+=3) for (a=ox;a;a--,out+=3){
+			out[0]=inp[0];
+			out[1]=inp[1];
+			out[2]=inp[2];
 		}
-		mem_free(ina);
+		mem_free(in);
 		return;
 	}
-	total=multiply_int(ix-1,ox-1);
-	if ((unsigned)y > MAXINT / 3 / sizeof(*col_buf)) overalloc();
-	col_buf=mem_alloc_mayfail(y*3*sizeof(*col_buf));
-	if (!col_buf) {
-		mem_free(ina);
+	multiply_int(ix-1,ox-1);
+
+	omp_start();
+#ifdef HAVE_OPENMP
+	use_omp = !OPENMP_NONATOMIC && !disable_openmp && (ox >= 24);
+#pragma omp parallel default(none) shared(col_buf) firstprivate(in,out,ix,ox,y,skip,oskip) if (use_omp)
+#endif
+	{
+		scale_t *thread_col_buf;
+		int alloc_size;
+		int out_idx;
+		if ((unsigned)y > (MAXINT - SMP_ALIGN + 1) / 3 / sizeof(*col_buf)) overalloc();
+		alloc_size = (int)(y*3*sizeof(*col_buf));
+		alloc_size = (alloc_size + SMP_ALIGN - 1) & ~(SMP_ALIGN - 1);
+#ifdef HAVE_OPENMP
+#pragma omp single
+#endif
+		{
+			int n_threads = omp_get_num_threads();
+			if (alloc_size > MAXINT / n_threads) overalloc();
+			col_buf=mem_alloc_mayfail(alloc_size * n_threads);
+		}
+		if (!col_buf) goto skip_omp;
+		thread_col_buf = (scale_t *)((char *)col_buf + alloc_size * omp_get_thread_num());
+		bias_buf_color(thread_col_buf, y, (scale_t)(ox - 1) / 2);
+#ifdef HAVE_OPENMP
+#pragma omp for nowait
+#endif
+		for (out_idx = 0; out_idx <= ox - 1; out_idx++) {
+			ulonglong out_pos, in_pos, in_end;
+			int in_idx;
+			out_pos = (ulonglong)out_idx * (ix - 1);
+			if (out_idx)
+				in_idx = (int)((out_pos - 1) / (ox - 1));
+			else
+				in_idx = 0;
+			in_pos = in_idx * (ox - 1);
+			in_end = in_pos + (ox - 1);
+			add_col_color(thread_col_buf, in + in_idx * 3, skip, y,
+				in_end - out_pos);
+			add_col_color(thread_col_buf, in + (in_idx + 1) * 3, skip, y,
+				out_pos - in_pos);
+			emit_and_bias_col_color(thread_col_buf, out + out_idx * 3, oskip, y, ox - 1);
+		}
+		skip_omp:;
+	}
+	omp_end();
+
+	mem_free(in);
+	if (col_buf) mem_free(col_buf);
+	else {
 		mem_free(out);
 		*outa = NULL;
-		return;
 	}
-	bias_buf_color(col_buf,y,(scale_t)(ox-1) / 2);
-	out_pos=0;
-	in_pos=0;
-	again:
-	in_begin=in_pos;
-	in_end=in_pos+ox-1;
-	add_col_color(col_buf,in,skip,y,
-		in_end-out_pos);
-	add_col_color(col_buf,in+3,skip,y,
-		out_pos-in_begin);
-	emit_and_bias_col_color(col_buf,out,oskip,y,ox-1);
-	out+=3;
-	out_pos+=ix-1;
-	if (out_pos>in_end){
-		in_pos=in_end;
-		in+=3;
-	}
-	if (out_pos>total){
-		mem_free(col_buf);
-		mem_free(ina);
-		return;
-	}
-	goto again;
 }
 
 /* Works for both enlarging and diminishing. Linear resample, no low pass.
@@ -528,72 +548,96 @@ static inline void scale_gray_horizontal(unsigned char *in, int ix, int y,
  * Frees ina and allocates outa.
  * If ox*3<=ix, and display_optimize, performs optimization for LCD.
  */
-static inline void scale_color_horizontal(unsigned short *ina, int ix, int y,
+static inline void scale_color_horizontal(unsigned short *in, int ix, int y,
 		unsigned short **outa, int ox)
 {
+#ifdef HAVE_OPENMP
+	int use_omp;
+#endif
 	scale_t *col_buf;
-	ulonglong total,out_pos,in_pos,in_begin,in_end,out_end;
 	unsigned skip=3*ix;
 	unsigned oskip=3*ox;
-	unsigned short *in, *out;
+	unsigned short *out;
 
-	if (!ina) {
+	if (!in) {
 		*outa = NULL;
 		return;
 	}
 
 	if (ix==ox){
-		*outa=ina;
+		*outa=in;
 		return;
 	}
 	if (ix<ox){
-		enlarge_color_horizontal(ina,ix,y,outa,ox);
+		enlarge_color_horizontal(in,ix,y,outa,ox);
 		return;
 	}
-	total=multiply_int(ix,ox);
+	multiply_int(ix,ox);
 	if (ox && (unsigned)ox * (unsigned)y / (unsigned)ox != (unsigned)y) overalloc();
 	if ((unsigned)ox * (unsigned)y > MAXINT / 3 / sizeof(*out)) overalloc();
 	out=mem_alloc_mayfail(sizeof(*out)*3*ox*y);
 	*outa=out;
 	if (!out) {
-		mem_free(ina);
+		mem_free(in);
 		return;
 	}
-	in=ina;
-	if ((unsigned)y > MAXINT / 3 / sizeof(*col_buf)) overalloc();
-	col_buf=mem_alloc_mayfail(y*3*sizeof(*col_buf));
-	if (!col_buf) {
-		mem_free(ina);
+
+	omp_start();
+#ifdef HAVE_OPENMP
+	use_omp = !OPENMP_NONATOMIC & !disable_openmp & (ox >= 24);
+#pragma omp parallel default(none) shared(col_buf) firstprivate(in,out,ix,ox,y,skip,oskip) if (use_omp)
+#endif
+	{
+		scale_t *thread_col_buf;
+		int alloc_size;
+		int out_idx;
+		if ((unsigned)y > (MAXINT - SMP_ALIGN + 1) / 3 / sizeof(*col_buf)) overalloc();
+		alloc_size = (int)(y*3*sizeof(*col_buf));
+		alloc_size = (alloc_size + SMP_ALIGN - 1) & ~(SMP_ALIGN - 1);
+#ifdef HAVE_OPENMP
+#pragma omp single
+#endif
+		{
+			int n_threads = omp_get_num_threads();
+			if (alloc_size > MAXINT / n_threads) overalloc();
+			col_buf = mem_alloc_mayfail(alloc_size * n_threads);
+		}
+		if (!col_buf) goto skip_omp;
+		thread_col_buf = (scale_t *)((char *)col_buf + alloc_size * omp_get_thread_num());
+		bias_buf_color(thread_col_buf, y, (scale_t)ix / 2);
+#ifdef HAVE_OPENMP
+#pragma omp for nowait
+#endif
+		for (out_idx = 0; out_idx < ox; out_idx++) {
+			ulonglong out_pos, out_end, in_pos;
+			int in_idx;
+			out_pos = (ulonglong)out_idx * ix;
+			out_end = out_pos + ix;
+			in_idx = (int)(out_pos / ox);
+			in_pos = in_idx * ox;
+			do {
+				ulonglong in_begin, in_end;
+				in_begin = in_pos;
+				in_end = in_pos + ox;
+				if (in_begin < out_pos) in_begin = out_pos;
+				if (in_end > out_end) in_end = out_end;
+				add_col_color(thread_col_buf, in + in_idx * 3, skip, y, in_end - in_begin);
+				in_idx++;
+				in_pos += ox;
+			} while (in_pos < out_end);
+
+			emit_and_bias_col_color(thread_col_buf, out + out_idx * 3, oskip, y, ix);
+		}
+		skip_omp:;
+	}
+	omp_end();
+
+	mem_free(in);
+	if (col_buf) mem_free(col_buf);
+	else {
 		mem_free(out);
 		*outa = NULL;
-		return;
 	}
-	bias_buf_color(col_buf,y,(scale_t)ix / 2);
-	out_pos=0;
-	in_pos=0;
-	again:
-	in_begin=in_pos;
-	in_end=in_pos+ox;
-	out_end=out_pos+ix;
-	if (in_begin<out_pos)in_begin=out_pos;
-	if (in_end>out_end)in_end=out_end;
-	add_col_color(col_buf,in,skip,y,in_end-in_begin);
-	in_end=in_pos+ox;
-	if (out_end>=in_end){
-		in_pos=in_end;
-		in+=3;
-	}
-	if (out_end<=in_end){
-			emit_and_bias_col_color(col_buf,out,oskip,y,ix);
-			out_pos=out_pos+ix;
-			out+=3;
-	}
-	if (out_pos==total) {
-		mem_free(ina);
-		mem_free(col_buf);
-		return;
-	}
-	goto again;
 }
 
 /* For magnification only. Does linear filtering. */
@@ -654,20 +698,22 @@ static inline void enlarge_gray_vertical(unsigned char *in, int x, int iy,
 
 /* For magnification only. Does linear filtering */
 /* We assume unsigned holds at least 32 bits */
-static inline void enlarge_color_vertical(unsigned short *ina, int x, int iy,
+static inline void enlarge_color_vertical(unsigned short *in, int x, int iy,
 	unsigned short **outa ,int oy)
 {
+#ifdef HAVE_OPENMP
+	int use_omp;
+#endif
 	scale_t *row_buf;
-	ulonglong total,out_pos,in_pos,in_begin,in_end;
-	unsigned short *out, *in;
+	unsigned short *out;
 
-	if (!ina) {
+	if (!in) {
 		*outa = NULL;
 		return;
 	}
 
 	if (iy==oy){
-		*outa=ina;
+		*outa=in;
 		return;
 	}
 	/* Rivendell */
@@ -676,50 +722,71 @@ static inline void enlarge_color_vertical(unsigned short *ina, int x, int iy,
 	out=mem_alloc_mayfail(sizeof(*out)*3*oy*x);
 	*outa=out;
 	if (!out) {
-		mem_free(ina);
+		mem_free(in);
 		return;
 	}
-	in=ina;
 	if (iy==1){
 		for (;oy;oy--){
-	       		memcpy(out,in,3*x*sizeof(*out));
-	       		out+=3*x;
+			memcpy(out,in,3*x*sizeof(*out));
+			out+=3*x;
 		}
-		mem_free(ina);
+		mem_free(in);
 		return;
 	}
-	total=multiply_int(iy-1,oy-1);
-	if ((unsigned)x > MAXINT / 3 / sizeof(*row_buf)) overalloc();
-	row_buf=mem_alloc_mayfail(x*3*sizeof(*row_buf));
-	if (!row_buf) {
-		mem_free(ina);
+	multiply_int(iy-1,oy-1);
+
+	omp_start();
+#ifdef HAVE_OPENMP
+	use_omp = (!OPENMP_NONATOMIC | !(x & 1)) & !disable_openmp & (oy >= 24);
+#pragma omp parallel default(none) shared(row_buf) firstprivate(in,out,x,iy,oy) if (use_omp)
+#endif
+	{
+		scale_t *thread_row_buf;
+		int alloc_size;
+		int out_idx;
+		if ((unsigned)x > (MAXINT - SMP_ALIGN + 1) / 3 / sizeof(*row_buf)) overalloc();
+		alloc_size = (int)(x*3*sizeof(*row_buf));
+		alloc_size = (alloc_size + SMP_ALIGN - 1) & ~(SMP_ALIGN - 1);
+#ifdef HAVE_OPENMP
+#pragma omp single
+#endif
+		{
+			int n_threads = omp_get_num_threads();
+			if (alloc_size > MAXINT / n_threads) overalloc();
+			row_buf=mem_alloc_mayfail(alloc_size * n_threads);
+		}
+		if (!row_buf) goto skip_omp;
+		thread_row_buf = (scale_t *)((char *)row_buf + alloc_size * omp_get_thread_num());
+		bias_buf_color(thread_row_buf,x,(scale_t)(oy-1) / 2);
+#ifdef HAVE_OPENMP
+#pragma omp for nowait
+#endif
+		for (out_idx = 0; out_idx <= oy - 1; out_idx++) {
+			ulonglong out_pos, in_pos, in_end;
+			int in_idx;
+			out_pos = (ulonglong)out_idx * (iy - 1);
+			if (out_idx)
+				in_idx = (int)((out_pos - 1) / (oy - 1));
+			else
+				in_idx = 0;
+			in_pos = in_idx * (oy - 1);
+			in_end = in_pos + oy - 1;
+			add_row_color(thread_row_buf, in + in_idx * 3 * x, x,
+				in_end - out_pos);
+			add_row_color(thread_row_buf, in + (in_idx + 1) * 3 * x, x,
+				out_pos - in_pos);
+			emit_and_bias_row_color(thread_row_buf, out + out_idx * 3 * x, x, oy - 1);
+		}
+		skip_omp:;
+	}
+	omp_end();
+
+	mem_free(in);
+	if (row_buf) mem_free(row_buf);
+	else {
 		mem_free(out);
 		*outa = NULL;
-		return;
 	}
-	bias_buf_color(row_buf,x,(scale_t)(oy-1) / 2);
-	out_pos=0;
-	in_pos=0;
-	again:
-	in_begin=in_pos;
-	in_end=in_pos+oy-1;
-	add_row_color(row_buf,in,x,
-		in_end-out_pos);
-	add_row_color(row_buf,in+3*x,x,
-		out_pos-in_begin);
-	emit_and_bias_row_color(row_buf,out,x,oy-1);
-	out+=3*x;
-	out_pos+=iy-1;
-	if (out_pos>in_end){
-		in_pos=in_end;
-		in+=3*x;
-	}
-	if (out_pos>total){
-		mem_free(ina);
-		mem_free(row_buf);
-		return;
-	}
-	goto again;
 }
 
 /* Both enlarges and diminishes. Linear filtering.
@@ -783,70 +850,91 @@ static inline void scale_gray_vertical(unsigned char *in, int x, int iy,
    We assume unsigned short can hold at least 16 bits.
    We assume unsigned holds at least 32 bits.
  */
-static inline void scale_color_vertical(unsigned short *ina, int x, int iy,
+static inline void scale_color_vertical(unsigned short *in, int x, int iy,
 	unsigned short **outa, int oy)
 {
+#ifdef HAVE_OPENMP
+	int use_omp;
+#endif
 	scale_t *row_buf;
-	ulonglong total,out_pos,in_pos,in_begin,in_end,out_end;
-	unsigned short *in, *out;
+	unsigned short *out;
 
-	if (!ina) {
+	if (!in) {
 		*outa = NULL;
 		return;
 	}
 
 	if (iy==oy){
-		*outa=ina;
+		*outa=in;
 		return;
 	}
 	if (iy<oy){
-		enlarge_color_vertical(ina,x,iy,outa,oy);
+		enlarge_color_vertical(in,x,iy,outa,oy);
 		return;
 	}
-	total=multiply_int(iy,oy);
+	multiply_int(iy,oy);
 	if (x && (unsigned)x * (unsigned)oy / (unsigned)x != (unsigned)oy) overalloc();
 	if ((unsigned)x * (unsigned)oy > MAXINT / 3 / sizeof(*out)) overalloc();
 	out=mem_alloc_mayfail(sizeof(*out)*3*oy*x);
 	*outa=out;
 	if (!out) {
-		mem_free(ina);
+		mem_free(in);
 		return;
 	}
-	in=ina;
-	if ((unsigned)x > MAXINT / 3 / sizeof(*row_buf)) overalloc();
-	row_buf=mem_alloc_mayfail(x*3*sizeof(*row_buf));
-	if (!row_buf) {
-		mem_free(ina);
+	omp_start();
+#ifdef HAVE_OPENMP
+	use_omp = (!OPENMP_NONATOMIC | !(x & 1)) & !disable_openmp & (oy >= 24);
+#pragma omp parallel default(none) shared(row_buf) firstprivate(in,out,x,iy,oy) if (use_omp)
+#endif
+	{
+		scale_t *thread_row_buf;
+		int alloc_size;
+		int out_idx;
+		if ((unsigned)x > (MAXINT - SMP_ALIGN + 1) / 3 / sizeof(*row_buf)) overalloc();
+		alloc_size = (int)(x*3*sizeof(*row_buf));
+		alloc_size = (alloc_size + SMP_ALIGN - 1) & ~(SMP_ALIGN - 1);
+#ifdef HAVE_OPENMP
+#pragma omp single
+#endif
+		{
+			int n_threads = omp_get_num_threads();
+			if (alloc_size > MAXINT / n_threads) overalloc();
+			row_buf = mem_alloc_mayfail(alloc_size * n_threads);
+		}
+		if (!row_buf) goto skip_omp;
+		thread_row_buf = (scale_t *)((char *)row_buf + alloc_size * omp_get_thread_num());
+		bias_buf_color(thread_row_buf,x,(scale_t)iy / 2);
+#ifdef HAVE_OPENMP
+#pragma omp for nowait
+#endif
+		for (out_idx = 0; out_idx < oy; out_idx++) {
+			ulonglong out_pos, out_end, in_pos;
+			int in_idx;
+			out_pos = (ulonglong)out_idx * iy;
+			out_end = out_pos + iy;
+			in_idx = (int)(out_pos / oy);
+			in_pos = in_idx * oy;
+			do {
+				ulonglong in_begin, in_end;
+				in_begin = in_pos;
+				in_end = in_pos + oy;
+				if (in_begin < out_pos) in_begin = out_pos;
+				if (in_end > out_end) in_end = out_end;
+				add_row_color(thread_row_buf, in + in_idx * 3 * x, x, in_end - in_begin);
+				in_idx++;
+				in_pos += oy;
+			} while (in_pos < out_end);
+			emit_and_bias_row_color(thread_row_buf, out + out_idx * 3 * x, x, iy);
+		}
+		skip_omp:;
+	}
+	omp_end();
+	mem_free(in);
+	if (row_buf) mem_free(row_buf);
+	else {
 		mem_free(out);
 		*outa = NULL;
-		return;
 	}
-	bias_buf_color(row_buf,x,(scale_t)iy / 2);
-	out_pos=0;
-	in_pos=0;
-	again:
-	in_begin=in_pos;
-	in_end=in_pos+oy;
-	out_end=out_pos+iy;
-	if (in_begin<out_pos)in_begin=out_pos;
-	if (in_end>out_end)in_end=out_end;
-	add_row_color(row_buf,in,x,in_end-in_begin);
-	in_end=in_pos+oy;
-	if (out_end>=in_end){
-		in_pos=in_end;
-		in+=3*x;
-	}
-	if (out_end<=in_end){
-			emit_and_bias_row_color(row_buf,out,x,iy);
-			out_pos=out_pos+iy;
-			out+=3*x;
-	}
-	if (out_pos==total){
-		mem_free(ina);
-		mem_free(row_buf);
-		return;
-	}
-	goto again;
 }
 
 
@@ -1114,9 +1202,9 @@ void agx_and_uc_32_to_48(unsigned short *dest,
 		r=fd_pow(r,red_gamma);
 		g=fd_pow(g,green_gamma);
 		b=fd_pow(b,blue_gamma);
-		ri=(unsigned)((r*65535)+0.5);
-		gi=(unsigned)((g*65535)+0.5);
-		bi=(unsigned)((b*65535)+0.5);
+		ri=(unsigned)((r*65535)+(float_double)0.5);
+		gi=(unsigned)((g*65535)+(float_double)0.5);
+		bi=(unsigned)((b*65535)+(float_double)0.5);
 		cmd_limit_16(ri);
 		cmd_limit_16(gi);
 		cmd_limit_16(bi);
@@ -1168,9 +1256,9 @@ void agx_and_uc_64_to_48(unsigned short *dest,
 		r=fd_pow(r,red_gamma);
 		g=fd_pow(g,green_gamma);
 		b=fd_pow(b,blue_gamma);
-		ri=(unsigned short)(r*65535+0.5);
-		gi=(unsigned short)(g*65535+0.5);
-		bi=(unsigned short)(b*65535+0.5);
+		ri=(unsigned short)(r*65535+(float_double)0.5);
+		gi=(unsigned short)(g*65535+(float_double)0.5);
+		bi=(unsigned short)(b*65535+(float_double)0.5);
 		cmd_limit_16(ri);
 		cmd_limit_16(gi);
 		cmd_limit_16(bi);
@@ -1249,17 +1337,17 @@ void agx_48_to_48(unsigned short *dest,
 		a=*src;
 		a*=inv_65535;
 		a=fd_pow(a,red_gamma);
-		dest[0]=(unsigned short)((a*65535)+0.5);
+		dest[0]=(unsigned short)((a*65535)+(float_double)0.5);
 		cmd_limit_16(dest[0]);
 		a=src[1];
 		a*=inv_65535;
 		a=fd_pow(a,green_gamma);
-		dest[1]=(unsigned short)((a*65535)+0.5);
+		dest[1]=(unsigned short)((a*65535)+(float_double)0.5);
 		cmd_limit_16(dest[1]);
 		a=src[2];
 		a*=inv_65535;
 		a=fd_pow(a,blue_gamma);
-		dest[2]=(unsigned short)((a*65535)+0.5);
+		dest[2]=(unsigned short)((a*65535)+(float_double)0.5);
 		cmd_limit_16(dest[2]);
 	}
 }
@@ -1295,17 +1383,17 @@ void agx_24_to_48(unsigned short *dest, const unsigned char *src, int
 		a=*src;
 		a*=inv_255;
 		a=fd_pow(a,red_gamma);
-		dest[0]=(unsigned short)((a*65535)+0.5);
+		dest[0]=(unsigned short)((a*65535)+(float_double)0.5);
 		cmd_limit_16(dest[0]);
 		a=src[1];
 		a*=inv_255;
 		a=fd_pow(a,green_gamma);
-		dest[1]=(unsigned short)((a*65535)+0.5);
+		dest[1]=(unsigned short)((a*65535)+(float_double)0.5);
 		cmd_limit_16(dest[1]);
 		a=src[2];
 		a*=inv_255;
 		a=fd_pow(a,blue_gamma);
-		dest[2]=(unsigned short)((a*65535)+0.5);
+		dest[2]=(unsigned short)((a*65535)+(float_double)0.5);
 		cmd_limit_16(dest[2]);
 	}
 }
@@ -1315,9 +1403,9 @@ void agx_24_to_48(unsigned short *dest, const unsigned char *src, int
  * We assume unsigned short holds at least 16 bits. */
 void make_gamma_table(struct cached_image *cimg)
 {
-	float_double rg=(float_double)(user_gamma/cimg->red_gamma);
-	float_double gg=(float_double)(user_gamma/cimg->green_gamma);
-	float_double bg=(float_double)(user_gamma/cimg->blue_gamma);
+	float_double rg=(float_double)((float_double)user_gamma/cimg->red_gamma);
+	float_double gg=(float_double)((float_double)user_gamma/cimg->green_gamma);
+	float_double bg=(float_double)((float_double)user_gamma/cimg->blue_gamma);
 	int a;
 	unsigned short *ptr_16;
 	unsigned short last_val;
@@ -1329,17 +1417,17 @@ void make_gamma_table(struct cached_image *cimg)
 		ptr_16=mem_alloc(768*sizeof(*(cimg->gamma_table)));
 		cimg->gamma_table=ptr_16;
 		for (a=0;a<256;a++,ptr_16++){
-			last_val = (unsigned short)(65535*fd_pow((float_double)a*inv_255,rg)+0.5);
+			last_val = (unsigned short)(65535*fd_pow((float_double)a*inv_255,rg)+(float_double)0.5);
 			cmd_limit_16(last_val);
 			*ptr_16 = last_val;
 		}
 		for (a=0;a<256;a++,ptr_16++){
-			last_val = (unsigned short)(65535*fd_pow((float_double)a*inv_255,gg)+0.5);
+			last_val = (unsigned short)(65535*fd_pow((float_double)a*inv_255,gg)+(float_double)0.5);
 			cmd_limit_16(last_val);
 			*ptr_16 = last_val;
 		}
 		for (a=0;a<256;a++,ptr_16++){
-			last_val = (unsigned short)(65535*fd_pow((float_double)a*inv_255,bg)+0.5);
+			last_val = (unsigned short)(65535*fd_pow((float_double)a*inv_255,bg)+(float_double)0.5);
 			cmd_limit_16(last_val);
 			*ptr_16 = last_val;
 		}
@@ -1355,21 +1443,21 @@ void make_gamma_table(struct cached_image *cimg)
 		cimg->gamma_table=ptr_16;
 		for (a=0;a<0x10000;a++,ptr_16++){
 			if (!x_slow_fpu || !(a & 0xff)) {
-				last_val = (unsigned short)(65535*fd_pow((float_double)a*inv_65535,rg)+0.5);
+				last_val = (unsigned short)(65535*fd_pow((float_double)a*inv_65535,rg)+(float_double)0.5);
 				cmd_limit_16(last_val);
 			}
 			*ptr_16 = last_val;
 		}
 		for (a=0;a<0x10000;a++,ptr_16++){
 			if (!x_slow_fpu || !(a & 0xff)) {
-				last_val = (unsigned short)(65535*fd_pow((float_double)a*inv_65535,gg)+0.5);
+				last_val = (unsigned short)(65535*fd_pow((float_double)a*inv_65535,gg)+(float_double)0.5);
 				cmd_limit_16(last_val);
 			}
 			*ptr_16 = last_val;
 		}
 		for (a=0;a<0x10000;a++,ptr_16++){
 			if (!x_slow_fpu || !(a & 0xff)) {
-				last_val = (unsigned short)(65535*fd_pow((float_double)a*inv_65535,bg)+0.5);
+				last_val = (unsigned short)(65535*fd_pow((float_double)a*inv_65535,bg)+(float_double)0.5);
 				cmd_limit_16(last_val);
 			}
 			*ptr_16 = last_val;
@@ -1394,7 +1482,7 @@ void agx_24_to_48_table(unsigned short *dest, const unsigned char *src, int
 unsigned char ags_8_to_8(unsigned char input, float gamma)
 {
 	const float_double inv_255=1/255.;
-	return 255*fd_pow((float_double)input*inv_255,gamma)+0.5;
+	return 255*fd_pow((float_double)input*inv_255,gamma)+(float_double)0.5;
 }
 #endif
 
@@ -1409,7 +1497,7 @@ unsigned short ags_8_to_16(unsigned char input, float gamma)
 	a*=inv_255;
 	a=fd_pow(a,gamma);
 	a*=65535;
-	retval = (unsigned short)(a+0.5);
+	retval = (unsigned short)(a+(float_double)0.5);
 	cmd_limit_16(retval);
 	return retval;
 }
@@ -1419,7 +1507,7 @@ unsigned short ags_8_to_16(unsigned char input, float gamma)
 unsigned char ags_16_to_8(unsigned short input, float gamma)
 {
 	const float_double inv_65535=(float_double)(1/65535.);
-	return (unsigned char)(fd_pow((float_double)input*inv_65535,gamma)*255+0.5);
+	return (unsigned char)(fd_pow((float_double)input*inv_65535,gamma)*255+(float_double)0.5);
 }
 
 /* Input is 0-65535 (16-bit). Output is 0-255 (8-bit)*/
@@ -1428,7 +1516,7 @@ unsigned short ags_16_to_16(unsigned short input, float gamma)
 	unsigned short retval;
 	const float_double inv_65535=(float_double)(1/65535.);
 
-	retval = (unsigned short)(65535*fd_pow((float_double)input*inv_65535,gamma)+0.5);
+	retval = (unsigned short)(65535*fd_pow((float_double)input*inv_65535,gamma)+(float_double)0.5);
 	cmd_limit_16(retval);
 	return retval;
 }
@@ -1664,7 +1752,7 @@ const unsigned char *png_data, int png_length, struct style *style)
 	for (y0=y;y0;y0--){
 		i2ptr[-1]=0;
 		memcpy(i2ptr,dptr,*x);
-		i2ptr[ix-1]=0;
+		i2ptr[*x]=0;
 		i2ptr+=ix;
 		dptr+=*x;
 	}
@@ -1677,12 +1765,12 @@ const unsigned char *png_data, int png_length, struct style *style)
 
 	/* Compute the matrix constants from contrast and sharpness */
 	conv0=(1+sharpness)*contrast;
-	conv1=(float)(-sharpness*0.25*contrast);
+	conv1=(float)(-sharpness*(float)0.25*contrast);
 
 	for (y0=y;y0;y0--){
 		for (x0=*x;x0;x0--){
 			/* Convolution */
-			c=(int)(((*i2ptr)*conv0)+i2ptr[-ix]*conv1+i2ptr[-1]*conv1+i2ptr[1]*conv1+i2ptr[ix]*conv1+0.5);
+			c=(int)(((*i2ptr)*conv0)+i2ptr[-ix]*conv1+i2ptr[-1]*conv1+i2ptr[1]*conv1+i2ptr[ix]*conv1+(float)0.5);
 			if (((unsigned)c)>=256) c=c<0?0:255;
 			*dptr=(unsigned char)c;
 			dptr++;
@@ -1749,9 +1837,9 @@ static struct font_cache_entry *supply_color_cache_entry(struct style *style, st
 	mix_two_colors(primary_data, found->bitmap.data,
 		found->bitmap.x*found->bitmap.y,
 		red,green,blue,
-		ags_8_to_16(style->r1,(float)(user_gamma/sRGB_gamma)),
-		ags_8_to_16(style->g1,(float)(user_gamma/sRGB_gamma)),
-		ags_8_to_16(style->b1,(float)(user_gamma/sRGB_gamma))
+		ags_8_to_16(style->r1,(float)((float)user_gamma/(float)sRGB_gamma)),
+		ags_8_to_16(style->g1,(float)((float)user_gamma/(float)sRGB_gamma)),
+		ags_8_to_16(style->b1,(float)((float)user_gamma/(float)sRGB_gamma))
 	);
 	if (display_optimize){
 		/* A correction for LCD */
@@ -2313,9 +2401,9 @@ long real_dip_get_color_sRGB(int rgb)
 	rgb = hack_rgb(rgb);
 
 	round_color_sRGB_to_48(&r,&g,&b,rgb);
-	r=ags_16_to_8(r,(float)(1/display_red_gamma));
-	g=ags_16_to_8(g,(float)(1/display_green_gamma));
-	b=ags_16_to_8(b,(float)(1/display_blue_gamma));
+	r=ags_16_to_8(r,(float)(1/(float)display_red_gamma));
+	g=ags_16_to_8(g,(float)(1/(float)display_green_gamma));
+	b=ags_16_to_8(b,(float)(1/(float)display_blue_gamma));
 	new_rgb=b|(g<<8)|(r<<16);
 	gamma_cache_rgb = rgb;
 	/* The get_color takes values with gamma of display_*_gamma */
@@ -2330,7 +2418,7 @@ void get_links_icon(unsigned char **data, int *width, int *height, int *skip, in
 {
 	struct bitmap b;
 	unsigned short *tmp1;
-	float g=(float)(user_gamma/sRGB_gamma);
+	float g=(float)((float)user_gamma/(float)sRGB_gamma);
 
 	b.x=48;
 	b.y=48;
